#include "stdafx.h"

#include "FrameCache.h"

#include "exception.h"

#include <algorithm>
#include <functional>
#include <fstream>

LONG align(LONG n, LONG alignment)
{
	if (alignment == 0) return n;
	return (n + (alignment - 1)) &~ (alignment - 1);
}

// Frame

void Frame::assignHeader(const BITMAPINFOHEADER& header)
{
	m_header = header;

	const RGBQUAD* colors = reinterpret_cast<const BITMAPINFO*>(&header)->bmiColors;
	std::copy(colors, colors + header.biClrUsed, m_colors);
}

void Frame::alloc(DWORD size)
{
	m_hMapping = ASSERT_NE(CreateFileMapping(INVALID_HANDLE_VALUE, LPSECURITY_ATTRIBUTES(),
		PAGE_READWRITE, 0, size, LPCTSTR()), HANDLE(), "CreateFileMapping");
	m_pData = ASSERT_NE(MapViewOfFile(m_hMapping, FILE_MAP_WRITE, 0, 0, size), LPVOID(), "MapViewOfFile");
}

Frame::Frame(const BITMAPINFOHEADER& header, int width, int height)
{
	assignHeader(header);

	m_header.biWidth = width;
	m_header.biHeight = height;
	m_header.biSizeImage = align(width * header.biBitCount / 8, 4) * height;

	alloc(m_header.biSizeImage);

	//m_handle = ASSERT_NE(CreateDIBSection(0, bitmapInfo(), DIB_RGB_COLORS, &m_pData, 0, 0), 
	//	HBITMAP(), "CreateDIBSection");
}

Frame::Frame(LPVOID pData)
{
	const BYTE* pStart = reinterpret_cast<const BYTE*>(pData);

	const BITMAPINFOHEADER* pbih = reinterpret_cast<const BITMAPINFOHEADER*>(pStart);
	assignHeader(*pbih);

	alloc(header()->biSizeImage);

	//m_handle = ASSERT_NE(CreateDIBSection(0, bitmapInfo(), DIB_RGB_COLORS, &m_pData, 0, 0), 
	//	HBITMAP(), "CreateDIBSection");

	const BYTE* pBits = pStart + pbih->biSize + sizeof(RGBQUAD) * pbih->biClrUsed;
	std::copy(pBits, pBits + pbih->biSizeImage, reinterpret_cast<BYTE*>(m_pData));
}

Frame::Frame(const BITMAPINFOHEADER& header, const Bitmap& bitmap)
{
	assignHeader(header);
	m_header.biWidth = bitmap.width();
	m_header.biHeight = bitmap.height();
	DWORD size = m_header.biSizeImage = 
		align(m_header.biWidth * m_header.biBitCount / 8, 4) * m_header.biHeight;

	alloc(size);

	//m_handle = ASSERT_NE(CreateDIBSection(0, bitmapInfo(), DIB_RGB_COLORS, &m_pData, 0, 0), 
	//	HBITMAP(), "CreateDIBSection");

	CDC dc;
	dc.CreateCompatibleDC(0);

	ASSERT_NE(GetDIBits(dc, bitmap.handle(), 
		0, bitmap.height(), 
		m_pData, bitmapInfo(), DIB_RGB_COLORS), 0, "GetDIBits");
}

Frame::Frame(const Frame& other)
{
	assignHeader(*other.header());
	alloc(m_header.biSizeImage);

	const BYTE* sourceData = reinterpret_cast<const BYTE*>(other.data());
	std::copy(sourceData, sourceData + m_header.biSizeImage,
		reinterpret_cast<BYTE*>(m_pData));
}

Frame::~Frame()
{
	//DeleteObject(m_handle);
	UnmapViewOfFile(m_pData);
	CloseHandle(m_hMapping);
}


LPBITMAPINFOHEADER Frame::header() const
{
	return const_cast<LPBITMAPINFOHEADER>(&m_header);
}

LPVOID Frame::data() const
{
	return m_pData;
}


// PFrame

void SeparateFields::operator()(BITMAPINFOHEADER& bih) const
{
	bih.biWidth = 2 * align(bih.biWidth, 32 / bih.biBitCount);
	bih.biHeight /= 2;
}


void PFrame::assignHeader(const BITMAPINFOHEADER* pHeader)
{
	m_header = *pHeader;
	const RGBQUAD* colors = reinterpret_cast<const BITMAPINFO*>(pHeader)->bmiColors;
	std::copy(colors, colors + pHeader->biClrUsed, m_colors);
}

const BITMAPINFO* PFrame::bitmapInfo() const
{
	return reinterpret_cast<const BITMAPINFO*>(&m_header);
}

PFrame::PFrame(Frame* frame /*= 0*/)
: Base(frame)
{
	if (*this)
	{
		assignHeader(get()->header());
		m_handle = ASSERT_NE(CreateDIBSection(0, bitmapInfo(), DIB_RGB_COLORS, &m_pData, 
			get()->hMapping(), 0), HBITMAP(), "CreateDIBSection");
	}
	else
	{
		ZeroMemory(&m_header, sizeof(BITMAPINFOHEADER));
		m_header.biSize = sizeof(BITMAPINFOHEADER);
		m_handle = 0;
	}
}

PFrame::PFrame(const PFrame& other)
: Base(other), m_header(other.m_header), 
	m_handle(other ? ASSERT_NE(CreateDIBSection(0, bitmapInfo(), DIB_RGB_COLORS, &m_pData, 
		get()->hMapping(), 0), HBITMAP(), "CreateDIBSection") : 0)
{
}

void PFrame::swap(PFrame& other)
{
	Base::swap(other);
	std::swap(m_pData, other.m_pData);
	std::swap(m_header, other.m_header);
	for (size_t i = 0; i < 256; ++i)
	{
		std::swap(m_colors[i], other.m_colors[i]);
	}
	std::swap(m_handle, other.m_handle);
}

PFrame& PFrame::operator=(const PFrame& other)
{
	PFrame(other).swap(*this);
	return *this;
}

PFrame PFrame::clone() const
{
	return PFrame(new Frame(**this));
}

const BITMAPINFOHEADER* PFrame::header() const
{
	return &m_header;
}

LPBITMAPINFOHEADER PFrame::header()
{
	return &m_header;
}

LPVOID PFrame::data() const
{
	return m_pData;
}

LONG PFrame::width() const
{
	return m_header.biWidth;
}

LONG PFrame::height() const
{
	return m_header.biHeight;
}

CSize PFrame::size() const
{
	return CSize(width(), height());
}

CRect PFrame::rect() const
{
	return CRect(CPoint(0, 0), size());
}

CPoint PFrame::fieldOrigin(LONG order) const
{
	return CPoint(order * m_header.biWidth / 2, 0);
}

SIZE_T PFrame::stride() const
{
	return align(m_header.biWidth * m_header.biBitCount / 8, 4);
}

int PFrame::lineFromY(int y) const
{
	return m_header.biHeight < 0 ? y : (m_header.biHeight - 1 - y);
}

SIZE_T PFrame::byteFromX(int x) const
{
	return x * m_header.biBitCount / 8;
}

LPVOID PFrame::ptr(int x, int y) const
{
	return LPBYTE(data()) + stride() * lineFromY(y) + byteFromX(x);
}

const RGBTRIPLE& PFrame::at(LONG x, LONG y) const
{
	return *reinterpret_cast<const RGBTRIPLE*>(ptr(x, y));
}
RGBTRIPLE& PFrame::at(LONG x, LONG y)
{
	return *reinterpret_cast<RGBTRIPLE*>(ptr(x, y));
}

void PFrame::copyFrom(const PFrame& source, CRect sourceRect, CRect targetRect)
{
	return copyFrom(source, sourceRect, targetRect, &memcpy);
}

void PFrame::copyFrom(const PBitmap& source)
{
	//CDC dc;
	//dc.CreateCompatibleDC(0);

	//BITMAPINFO bi;
	//bi.bmiHeader = *header();

	//ASSERT_NE(GetDIBits(dc, source->handle(), 0, height(), data(), &bi, DIB_RGB_COLORS), 0, "GetDIBits");

	PFrame sourceFrame(new Frame(*header(), *source));
	copyFrom(sourceFrame, sourceFrame.rect(), sourceFrame.rect());
}



// AviFile

AviFile::AviFile()
	: m_pAviFile(0), m_pAviStream(0), m_pGetFrame(0), 
	m_length(0), m_millifps(0)
{
}

AviFile::~AviFile()
{
	close();
}

void AviFile::open(LPCTSTR lpszFilename)
{
	close();

	ASSERT_SUCCESS(AVIFileOpen(&m_pAviFile, lpszFilename, OF_READ | OF_SHARE_DENY_WRITE, 0), "AVIFileOpen");
	ASSERT_SUCCESS(AVIFileGetStream(m_pAviFile, &m_pAviStream, streamtypeVIDEO, 0), "AVIFileGetStream");

	LONG formatSize = 0;
	ASSERT_SUCCESS(AVIStreamFormatSize(m_pAviStream, 0, &formatSize), "AVIStreamFormatSize");
	if (formatSize < sizeof(BITMAPINFOHEADER)) throw std::runtime_error("AviFile::open: Frame format is too small");

	std::vector<BYTE> format(formatSize, 0);
	ASSERT_SUCCESS(AVIStreamReadFormat(m_pAviStream, 0, &format[0], &formatSize), "AVIStreamReadFormat");

	LPBITMAPINFOHEADER pHeader = reinterpret_cast<LPBITMAPINFOHEADER>(&format[0]);
	pHeader->biBitCount = 24;
	pHeader->biCompression = BI_RGB;
	pHeader->biSizeImage = align(pHeader->biWidth * pHeader->biBitCount / 8, 4) * pHeader->biHeight;

	m_pGetFrame = ASSERT_NE(AVIStreamGetFrameOpen(m_pAviStream, pHeader), 
		PGETFRAME(), "AVIStreamGetFrameOpen");
	m_length = ASSERT_NE(AVIStreamLength(m_pAviStream), -1, "AVIStreamLength");

	AVIFILEINFO fi;
	ASSERT_SUCCESS(AVIFileInfo(m_pAviFile, &fi, sizeof(fi)), "AVIFileInfo");
	m_millifps = fi.dwRate * 1000 / fi.dwScale;
	m_width = fi.dwWidth;
	m_height = fi.dwHeight;
}

void AviFile::close()
{
	m_length   = 0;
	m_millifps = 0;
	m_width    = 0;
	m_height   = 0;
	m_frames.clear();
	if (m_pGetFrame)  { AVIStreamGetFrameClose(m_pGetFrame); m_pGetFrame = 0; }
	if (m_pAviStream) { AVIStreamRelease(m_pAviStream); m_pAviStream = 0; }
	if (m_pAviFile)   { AVIFileRelease(m_pAviFile); m_pAviFile = 0; }
}

PFrame AviFile::operator[](LONG frame) const
{
	Frames::const_iterator it = m_frames.find(frame);
	if (it != m_frames.end()) return it->second;
	if (!m_pGetFrame) return PFrame();
	PFrame result(new Frame(AVIStreamGetFrame(m_pGetFrame, frame)));
	if (result) m_frames.insert(std::make_pair(frame, result));
	return result;
}

PFrame AviFile::getUncached(LONG frame) const
{
	if (!m_pGetFrame) return PFrame();
	return PFrame(new Frame(AVIStreamGetFrame(m_pGetFrame, frame)));
}

void AviFile::erase(LONG frame)
{
	m_frames.erase(frame);
}

LONG AviFile::size() const
{
	return m_length;
}

LONG AviFile::millifps() const
{
	return m_millifps;
}

LONG AviFile::width() const
{
	return m_width;
}

LONG AviFile::height() const
{
	return m_height;
}

CRect AviFile::frameSize() const
{
	return CRect(0, 0, width(), height());
}

Bitmap::Bitmap(LPCSTR lpszFileName)
{
	ASSERT_SUCCESS(m_image.Load(lpszFileName), "CImage::Load");
}

HBITMAP Bitmap::handle() const
{
	//return m_handle;
	return m_image;
}

CSize Bitmap::size() const
{
	//CSize result;
	//ASSERT_NE(GetBitmapDimensionEx(m_handle, &result), 0, "GetBitmapDimensionEx");
	//return result;
	return CSize(width(), height());
}

LONG Bitmap::width() const
{
	//return size().cx;
	return m_image.GetWidth();
}

LONG Bitmap::height() const
{
	//return size().cy;
	return m_image.GetHeight();
}

CRect Bitmap::rect() const
{
	return CRect(CPoint(0, 0), size());
}


PBitmap ImageCache::operator[](const std::string& fileName) const
{
	Images::const_iterator it = m_images.find(fileName);
	if (m_images.end() != it) return it->second;

	PBitmap result(new Bitmap(fileName.c_str()));
	m_images.insert(std::make_pair(fileName, result));

	return result;
}
