#include "stdafx.h"

#include "FrameRenderer.h"

#include "FrameOperation.h"
#include "ViewMode.h"
#include "TemporaryFile.h"

#include "FramePatch.h"
#include "FieldPatch.h"
#include "FrameImage.h"
#include "FieldRebuild.h"
#include "MotionInterpolate.h"

#include "exception.h"

#include <gdiplus.h>

#include <afxwin.h>

#include <algorithm>
#include "algorithms.h"
#include <functional>
#include "sgi_functional.h"

#include <fstream>
#include <ostream>

class GdiplusStartupShutdown
{
private:
	ULONG m_token;
public:
	GdiplusStartupShutdown()
	{
		Gdiplus::GdiplusStartupInput input;
		ASSERT_GDIP(Gdiplus::GdiplusStartup(&m_token, &input, 0), "GdiplusStartup");
	}
	~GdiplusStartupShutdown()
	{
		Gdiplus::GdiplusShutdown(m_token);
	}
};

GdiplusStartupShutdown g_gdiplusStartupShutdown;

class setTextColor
{
private:
	HDC m_hDC;
	COLORREF m_oldColor;
public:
	setTextColor(HDC hDC, COLORREF color)
		: m_hDC(hDC), m_oldColor(ASSERT_NE(SetTextColor(hDC, color), CLR_INVALID, "SetTextColor"))
	{
	}

	~setTextColor()
	{
		SetTextColor(m_hDC, m_oldColor);
	}
};

class setBkMode
{
private:
	HDC m_hDC;
	int m_oldMode;
public:
	setBkMode(HDC hDC, int mode)
		: m_hDC(hDC), m_oldMode(ASSERT_NE(SetBkMode(hDC, mode), 0, "SetBkMode"))
	{
	}

	~setBkMode()
	{
		SetBkMode(m_hDC, m_oldMode);
	}
};

void drawOutlinedText(CDC* pDC, COLORREF textColor, COLORREF outlineColor,
											const CString& string, CRect rect, UINT format)
{
	setBkMode sbm(*pDC, TRANSPARENT);

	setTextColor stcOutline(*pDC, outlineColor);

	ASSERT_NE(pDC->DrawText(string, rect + CPoint(-1, -1), format), 0, "DrawText(-1, -1)");
	ASSERT_NE(pDC->DrawText(string, rect + CPoint( 1, -1), format), 0, "DrawText( 1, -1)");
	ASSERT_NE(pDC->DrawText(string, rect + CPoint(-1,  1), format), 0, "DrawText(-1,  1)");
	ASSERT_NE(pDC->DrawText(string, rect + CPoint( 1,  1), format), 0, "DrawText( 1,  1)");

	setTextColor stcText(*pDC, textColor);

	ASSERT_NE(pDC->DrawText(string, rect, format), 0, "DrawText");
}



FrameRenderer::FrameRenderer(CDC* pDC,
														 CVclipDoc& document,
														 const AviFile& aviFile, const ImageCache& images,
														 bool evenCoords)
: m_pDC(pDC), m_document(&document), m_aviFile(&aviFile), m_images(&images), m_evenCoords(evenCoords)
{
}

PFrame FrameRenderer::render(LONG frameNumber) const
{
	return render(frameNumber, m_document->end());
}

PFrame FrameRenderer::render(LONG frameNumber, CVclipDoc::iterator end) const
{
	if (canRenderDirectly(frameNumber, end))
	{
		return renderDirectly(frameNumber, end);
	}
	else
	{
		return renderViaAvisynth(frameNumber, end);
	}
}

class CanRenderDirectly : public std::unary_function<PFrameOperation, bool>,
	public IOperationVisitor
{
private:
	const FrameRenderer* m_renderer;
	CVclipDoc* m_document;
	CVclipDoc::iterator m_end;
	mutable bool m_canDraw;
public:
	CanRenderDirectly(const FrameRenderer& renderer, CVclipDoc& document,
		CVclipDoc::iterator end)
	: m_renderer(&renderer), m_document(&document), m_end(end)
	{
	}

	bool operator()(const PFrameOperation& op)
	{
		op.accept(*this);
		return m_canDraw;
	}

	void operator()(FramePatch& patch)
	{
		m_canDraw = m_renderer->canRenderDirectly(patch.sourceFrame(), m_end);
	}
	void operator()(FieldPatch& patch)
	{
		m_canDraw = m_renderer->canRenderDirectly(patch.sourceFrame(), m_end);
	}
	void operator()(FrameImage& /*image*/)
	{
		m_canDraw = true;
	}
	void operator()(FieldRebuild& rebuild)
	{
		m_canDraw =
			m_renderer->canRenderDirectly(rebuild.targetFrame() - 1, m_end) &&
			m_renderer->canRenderDirectly(rebuild.targetFrame() + 1, m_end);
	}
	void operator()(MotionInterpolate& /*inter*/)
	{
		m_canDraw = false;
	}
};

bool FrameRenderer::canRenderDirectly(LONG frameNumber, CVclipDoc::iterator end) const
{
	for (CVclipDoc::iterator it = m_document->begin(); it != end; ++it)
	{
		if ((**it).targetFrame() == frameNumber &&
			!CanRenderDirectly(*this, *m_document, it)(*it))
		{
			return false;
		}
	}
	return true;
}

class DrawOperationInFrame : public std::unary_function<PFrameOperation, void>,
	public IOperationVisitor
{
private:
	CDC* m_pDC;
	PFrame* m_pFrame;
	//const AviFile* m_pAviFile;
	const ImageCache* m_pImageCache;
	const FrameRenderer* m_pRenderer;
	CVclipDoc::iterator m_end;
public:
	DrawOperationInFrame(CDC* pDC, PFrame& frame/*, const AviFile& aviFile*/, const ImageCache& imageCache,
		const FrameRenderer& renderer)
		: m_pDC(pDC), m_pFrame(&frame)/*, m_pAviFile(&aviFile)*/, m_pImageCache(&imageCache),
		m_pRenderer(&renderer)
	{
	}

	void operator()(CVclipDoc::iterator it)
	{
		m_end = it;
		it->accept(*this);
	}

	void operator()(FramePatch& patch)
	{
		PFrame resizedPatch(new Frame(*m_pFrame->header(), patch.targetWidth(), patch.targetHeight()));

		{
			PFrame sourceFrame = m_pRenderer->renderDirectly(patch.sourceFrame(), m_end);

			CDC dc;
			ASSERT_NE(dc.CreateCompatibleDC(0), FALSE, "CreateCompatibleDC");
			dc.SelectObject(resizedPatch.handle());

			CDC srcDc;
			ASSERT_NE(srcDc.CreateCompatibleDC(0), FALSE, "CreateCompatibleDC");
			srcDc.SelectObject(sourceFrame.handle());

			ASSERT_NE(dc.StretchBlt(0, 0, patch.targetWidth(), patch.targetHeight(),
				&srcDc, patch.sourceX(), patch.sourceY(), patch.sourceWidth(), patch.sourceHeight(),
				SRCCOPY), FALSE, "StretchBlt");
			GdiFlush();
		}

		m_pFrame->copyFrom(
			resizedPatch,
			resizedPatch.rect(),
			patch.targetRect(),
			*patch.mode());
	}

	void operator()(FieldPatch& patch)
	{
		CRect targetRect = fieldMode.viewFromFrame(patch.targetRect()) + CPoint(patch.targetOrder() * m_pFrame->width(), 0);
		CSize targetSize = targetRect.Size();
		CRect resizedRect(CPoint(0, 0), targetSize);

		PFrame resizedPatch(new Frame(*m_pFrame->header(), 
			targetSize.cx, targetSize.cy));

		{
			PFrame sourceFrame(m_pRenderer->renderDirectly(patch.sourceFrame(), m_end), SeparateFields());

			CDC dc;
			ASSERT_NE(dc.CreateCompatibleDC(0), FALSE, "CreateCompatibleDC");
			dc.SelectObject(resizedPatch.handle());

			CDC srcDc;
			ASSERT_NE(srcDc.CreateCompatibleDC(0), FALSE, "CreateCompatibleDC");
			srcDc.SelectObject(sourceFrame.handle());

			CRect sourceRect = fieldMode.viewFromFrame(patch.sourceRect()) + CPoint(patch.sourceOrder() * m_pFrame->width(), 0);
			
			ASSERT_NE(dc.StretchBlt(0, 0, targetSize.cx, targetSize.cy,
				&srcDc, sourceRect.left, sourceRect.top, sourceRect.Width(), sourceRect.Height(),
				SRCCOPY), FALSE, "StretchBlt");
			GdiFlush();
		}

		PFrame(*m_pFrame, SeparateFields()).copyFrom(resizedPatch,
			resizedRect,
			targetRect,
			*patch.mode());
	}

	void operator()(FrameImage& image)
	{
		PBitmap bitmap = (*m_pImageCache)[image.fileName()];
		m_pFrame->copyFrom(bitmap);
	}

	void operator()(FieldRebuild& rebuild)
	{
		PFrame fields(*m_pFrame, SeparateFields());
		PFrame previous(m_pRenderer->renderDirectly(rebuild.targetFrame() - 1, m_end), SeparateFields());
		PFrame next    (m_pRenderer->renderDirectly(rebuild.targetFrame() + 1, m_end), SeparateFields());

		CRect fieldRect(0, 0, m_pFrame->width(), fields.height());

		fields.copyFrom(previous, fieldRect + previous.fieldOrigin(0), fieldRect + fields.fieldOrigin(0));
		fields.copyFrom(next,     fieldRect + next    .fieldOrigin(1), fieldRect + fields.fieldOrigin(1));
	}

	void operator()(MotionInterpolate& rebuild)
	{
		CDC dc;
		ASSERT_NE(dc.CreateCompatibleDC(0), FALSE, "CreateCompatibleDC");

		CBitmap bitmap;
		ASSERT_NE(bitmap.CreateCompatibleBitmap(m_pDC, m_pFrame->width(), m_pFrame->height()), FALSE, "CreateCompatibleBitmap");
		ASSERT_NE(SetDIBits(*m_pDC, bitmap, 0, m_pFrame->height(), m_pFrame->data(),
			reinterpret_cast<LPBITMAPINFO>(m_pFrame->header()), DIB_RGB_COLORS), 0, "SetDIBits");

		CFont font;
		ASSERT_NE(font.CreateFont(
			0, 0, // width, height
			0, 0, // escapement, orientation
			FW_BOLD, FALSE, FALSE, 0, // weight, italic, underline, strikeout
			ANSI_CHARSET,
			OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
			DEFAULT_PITCH,
			"MS Shell Dlg"), FALSE, "CreateFont");

		ASSERT_NE(dc.SelectObject(bitmap), HGDIOBJ(), "SelectObject(bitmap)");
		ASSERT_NE(dc.SelectObject(font), HGDIOBJ(), "SelectObject(font)");

		CRect textRect = m_pFrame->rect();
		textRect.DeflateRect(10, 10);

		drawOutlinedText(&dc, RGB(255, 255, 0), RGB(0, 0, 0),
			"This frame (or part thereof) will be motion-interpolated.\n"
			"Vclip doesnt know how to display motion-interpolated frames, "
			"but can call upon the powers of Avisynth to render them. "
			"Select View | Accurate Rendering if you wish to see the exact "
			"Avisynth output. But please be aware that it may be slow.",
			textRect, DT_LEFT | DT_TOP | DT_NOPREFIX | DT_WORDBREAK);

		ASSERT_NE(GetDIBits(*m_pDC, bitmap, 0, m_pFrame->height(), m_pFrame->data(),
			reinterpret_cast<LPBITMAPINFO>(m_pFrame->header()), DIB_RGB_COLORS), 0, "GetDIBits");
	}
};

PFrame FrameRenderer::renderDirectly(LONG frameNumber, CVclipDoc::iterator end) const
{
	PFrame frame = m_aviFile->getUncached(frameNumber);
	yk::for_each_if(m_document->begin(), end,
		sgi::compose1(std::bind2nd(std::equal_to<LONG>(), frameNumber), TargetFrame()),
		DrawOperationInFrame(m_pDC, frame, *m_images, *this));
	return frame;
}

PFrame FrameRenderer::renderViaAvisynth(LONG frameNumber, CVclipDoc::iterator end) const
{
	TemporaryFilename avsName(".avs");

	std::ofstream avs(avsName.c_str());
	Compiler compiler(avs, m_evenCoords, m_aviFile->frameSize());
	compiler.source(m_document->source());
	std::for_each(m_document->begin(), end, compiler);
	compiler.end(frameNumber);
	avs.close();

	AviFile avsAviFile;
	avsAviFile.open(avsName.c_str());
	return avsAviFile.getUncached(frameNumber).clone();
}
